package ru.dbotthepony.kstarbound.server.network

import io.netty.channel.ChannelHandlerContext
import it.unimi.dsi.fastutil.objects.Object2ObjectOpenHashMap
import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kstarbound.client.network.packets.ForgetChunkPacket
import ru.dbotthepony.kstarbound.client.network.packets.InitialChunkDataPacket
import ru.dbotthepony.kstarbound.client.network.packets.SpawnWorldObjectPacket
import ru.dbotthepony.kstarbound.network.Connection
import ru.dbotthepony.kstarbound.network.ConnectionSide
import ru.dbotthepony.kstarbound.network.ConnectionType
import ru.dbotthepony.kstarbound.network.IServerPacket
import ru.dbotthepony.kstarbound.network.packets.HelloPacket
import ru.dbotthepony.kstarbound.server.StarboundServer
import ru.dbotthepony.kstarbound.server.world.ServerWorld
import ru.dbotthepony.kstarbound.world.ChunkPos
import ru.dbotthepony.kvector.vector.Vector2d
import java.util.*

class ServerConnection(val server: StarboundServer, type: ConnectionType) : Connection(ConnectionSide.SERVER, type, UUID(0L, 0L)) {
	var world: ServerWorld? = null
		set(value) {
			field = value
			needsToRecomputeTrackedChunks = true
		}

	var trackedPosition: Vector2d = Vector2d.ZERO
		set(value) {
			if (field != value) {
				field = value
				needsToRecomputeTrackedChunks = true
			}
		}

	var trackedPositionChunk: ChunkPos = ChunkPos.ZERO
		private set

	var trackedChunksWidth = 4
		set(value) {
			if (field != value) {
				field = value
				needsToRecomputeTrackedChunks = true
			}
		}

	var trackedChunksHeight = 4
		set(value) {
			if (field != value) {
				field = value
				needsToRecomputeTrackedChunks = true
			}
		}

	private val tickets = Object2ObjectOpenHashMap<ChunkPos, ServerWorld.ITicket>()
	private val sentChunks = ObjectOpenHashSet<ChunkPos>()

	private var needsToRecomputeTrackedChunks = true

	private fun recomputeTrackedChunks() {
		val world = world ?: return
		val trackedPositionChunk = world.geometry.chunkFromCell(trackedPosition)
		needsToRecomputeTrackedChunks = false
		if (trackedPositionChunk == this.trackedPositionChunk) return

		val tracked = ObjectOpenHashSet<ChunkPos>()

		for (x in -trackedChunksWidth .. trackedChunksWidth) {
			for (y in -trackedChunksHeight .. trackedChunksHeight) {
				tracked.add(world.geometry.wrap(trackedPositionChunk + ChunkPos(x, y)))
			}
		}

		val itr = tickets.entries.iterator()

		for ((pos, ticket) in itr) {
			if (pos !in tracked) {
				ticket.cancel()
				itr.remove()
			}
		}

		for (pos in tracked) {
			if (pos !in tickets) {
				tickets[pos] = world.permanentChunkTicket(pos)
			}
		}
	}

	fun tick() {
		val world = world

		if (world == null) {
			tickets.values.forEach { it.cancel() }
			tickets.clear()
			sentChunks.clear()
			return
		}

		if (needsToRecomputeTrackedChunks) {
			recomputeTrackedChunks()
		}

		for (pos in tickets.keys) {
			if (pos !in sentChunks) {
				val chunk = world.chunkMap[pos]

				if (chunk != null) {
					send(InitialChunkDataPacket(chunk))

					chunk.objects.forEach {
						send(SpawnWorldObjectPacket(it.serialize()))
					}

					sentChunks.add(pos)
				}
			}
		}

		val itr = sentChunks.iterator()

		for (pos in itr) {
			if (pos !in tickets) {
				send(ForgetChunkPacket(pos))
				itr.remove()
			}
		}
	}

	override fun channelRead(ctx: ChannelHandlerContext, msg: Any) {
		if (msg is IServerPacket) {
			try {
				msg.play(this)
			} catch (err: Throwable) {
				LOGGER.error("Failed to read serverbound packet $msg", err)
				disconnect(err.toString())
			}
		} else {
			LOGGER.error("Unknown serverbound packet type $msg")
			disconnect("Unknown serverbound packet type $msg")
		}
	}

	override fun inGame() {
		server.playerInGame(this)
	}

	override fun onHelloReceived(helloPacket: HelloPacket) {
	}

	companion object {
		private val LOGGER = LogManager.getLogger()
	}
}
